from .addPriorCount import addPriorCount
from .adjustedProfileLik import adjustedProfileLik
from .aveLogCPM import aveLogCPM
from .DGEGLM import DGEGLM
from .DGEList import DGEList
from .dispCoxReid import dispCoxReid
from .dispCoxReidInterpolateTagwise import dispCoxReidInterpolateTagwise
from .estimateGLMCommonDisp import estimateGLMCommonDisp
from .estimateGLMTagwiseDisp import estimateGLMTagwiseDisp
from .glmFit import glmFit
from .maximizeInterpolant import maximizeInterpolant
from .mglmLevenberg import mglmLevenberg
from .mglmOneGroup import mglmOneGroup
from .mglmOneWay import mglmOneWay, designAsFactor
from .movingAverageByCol import movingAverageByCol
from .nbinomDeviance import nbinomDeviance
from .predFC import predFC
from .splitIntoGroups import splitIntoGroups
from .stats import pnbinom, qnbinom
from .systematicSubset import systematicSubset
from .validDGEList import validDGEList
from .edgepy_cpp import *
